#!/usr/bin/perl -w
#
#   Copyright 2012 Andrii Golovynskyi
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

use Config::General;

use Cwd 'abs_path';
use File::Basename;

# Global variable 
our $prefix;
our $cmd_prefix;
our $cluster;
our $mpi_launcher;
our $queues;

my %config;

$ENV{LANG} = "C";

# Overriding config for remote cluster
if ((defined $cmd_prefix) && ($cmd_prefix ne "") &&
    (defined $cluster) && ($cluster ne "")) {
  $config{PREFIX} = "";
  $config{CLUSTER} = $cluster;
  $config{SLURM_MPI_LAUNCHER} = $mpi_launcher if defined $mpi_launcher;
} else {
  # Begin parsing config
  my $configfile = new Config::General("$prefix/etc/slurm.conf");
  %config = $configfile->getall;
  $cmd_prefix = "";
  $config{PREFIX} .= '/';
}

use Date::Parse;

require "$prefix/libexec/modules/utils.pl";
require "$prefix/libexec/modules/makefile.pl";

# Equvivalents of scontrol show hostnames
sub slurm_expand_nodes($){
  my $nodes = shift;
  my $enodes = "";
  $nodes =~ s/,([a-zA-Z])/ $1/;
  foreach my $node (split(" ",$nodes)){

    if( $node =~ m/([a-zA-Z]*)\[([0-9\-,]*)\]/ ){
      my $name = $1;
      my $list = $2;

      foreach my $element (split(",",$list)){
        if($element =~ /(\d*)-(\d*)/){
          my $start=$1;
          my $end=$2;
          for (my $i=$start;$i<=$end;$i++){
            $enodes .= $name.$i.", ";
          }
        } else {
          $enodes .= $name.$element.", ";
        }
      }
    } else {
      $enodes .= $node . ", ";
    }
  }
  chop $enodes;
  chop $enodes;
#    $enodes =~ s/, $//;
  return $enodes;
}

# --------------------------- Functions ---------------------------------------------------

sub launch {
  my ($J,$d,$a,$p,$D,$Djid,$n,$t,$r,$g,$ppn) = @_;
  my $slurmlauncher;
  if ( $r eq "mpi" ) {
    $slurmlauncher= $config{SLURM_MPI_LAUNCHER};
  } elsif ($r eq "none" ) {
    $slurmlauncher= "";
  } else { 
   print ("Unknown Runtime: $r\n");
   exit;
  }

  # Running task

  system ("mkdir -p $d/$J.out");

  my $dependstring="";
  my $res;
  my $ppn_str = "";
  my $g_str = "";
  $g_str = $g if defined $g;  

  if ((defined $ppn) && (int($ppn) > 0)) {
   $ppn_str = "--extra-node-info=*:$ppn"; 
  }

  if ((defined $D) && ($D eq  "afterok")) {
    $dependstring="--dependency=afterok:$Djid";
  } elsif ((defined $D) && ($D eq "afterfail")) {
    $dependstring="--dependency=afterfail:$Djid";
  }
  
  my $cmd_str = "cd $d && $config{PREFIX}sbatch -p $p $dependstring -n $n $ppn_str -t $t -J $J -o $d/$J.out/$J.out -e $d/$J.out/$J.err $slurmlauncher $a $g_str 2>&1";
  my %res;
  $res{cmd} = $cmd_str;
  $res{out} = ex($cmd_str);
  ($res{jobid}) = $res{out} =~ /(\d+)/;
  return \%res;
}


sub cancel {
# Cancel job "j", "a" -- administrator mode (unused)
  my ($j,$a) = @_;
  my $res;
  if ($a == 1 && ($cmd_prefix eq "")) {
    $admprefix="ssh root@".$config{ACCESS_SERVER};
  }
  $res = ex("$admprefix $config{PREFIX}scancel $j 2>&1");
  return $res;
}


sub resources {
# Get resources using sinfo 
# parameter detail - detailed view
  my ($detail) = @_;
  my @part_l;
  my $slurmprefix= $config{SLURM_PREFIX};
  my @res;
  if (defined $detail) {
    @res = ex("$config{PREFIX}sinfo -h -o \"\%P;\%a;\%N;\%t;\%D;\%E;\%s;\%c;\%l;\%m;\%C;\%f\"");
  } else {
    @res = ex("$config{PREFIX}sinfo -h -o \"\%P;\%a;\%N;\%t;\%D;\%l;\%C;\%f\"");
  }
  foreach my $l (@res) {
    $l =~ s/\n//;
    my %p;
    if (defined $detail) {
      ($p{P},$p{a},$p{N},$p{t},$p{D},$p{E},$p{s},$p{c},$p{l},$p{m},$p{C},$p{f}) = split(/;/,$l);
    } else {
      ($p{P},$p{a},$p{N},$p{t},$p{D},$p{l},$p{C},$p{f}) = split(/;/,$l);
    }
    delete $p{f} if ($p{f} eq "(null)");
    delete $p{E} if ( defined $p{E} && $p{E} eq "none");
    $p{C} = (split(/\//,$p{C}))[3];
    $p{l} =~ s/-/ /;
    $p{l} = str2sec($p{l}) || "U";
    $p{N} = slurm_expand_nodes($p{N}) if ($p{N} ne "");
    $p{H} = $config{CLUSTER};
    push @part_l, \%p;
  }
  return \@part_l;  
}


sub queue {
# Get queue state using sinfo or qstat
# parameter detail - detailed view
  my ($detail) = @_; 
  my $d = "";
  $d = "-t ALL" if (defined $detail && $detail == 1);
  my @res = ex("$config{PREFIX}squeue -h $d -o \"\%i;\%P;\%j;\%u;\%t;\%M;\%D;\%r;\%C;\%b;\%e;\%E;\%N\"");
  my @log = ex("$config{PREFIX}scontrol show jobs -o");
  my $log = parse_log_submittime(@log);

  my @queue;
  foreach my $l (@res) {
    $l =~ s/\n//;
    my %job;
    ($job{'i'},$job{'P'},$job{'j'}, $job{'u'},$job{'t'},$job{'M'},$job{'D'},
      $job{'r'},$job{'C'},$job{'b'},$job{'e'},$job{'E'},$job{'N'}) = split(/;/,$l);
    delete $job{'r'} if ($job{'r'} eq "None");
    delete $job{'E'} if ($job{'E'} eq "");
    $job{'N'} = slurm_expand_nodes($job{'N'}) if ("$job{'N'}" ne "");
    $job{'b'} =~ s/-/ / if ( $job{'b'} =~ m/\d+\/\d+-\d+:\d+:\d+/);
    $job{'e'} =~ s/-/ / if( $job{'e'} =~ m/\d+\/\d+-\d+:\d+:\d+/);
    $job{'e'} = str2time($job{'e'}) || "Unknown";
    $job{'b'} = str2time($job{'b'}) || "Unknown";
    $job{'M'} = str2sec($job{'M'}) || 0;
    $job{'H'} = $config{CLUSTER};

    my $id = $job{'i'};
#print "JID: $id, $log->{$id}->{S}\n";
    my $st = $log->{$job{'i'}}->{S};
#    $job{'S'} = $st;
    if ( ($job{'b'} ne "Unknown") && ($job{'b'} > 0) && $st > 0 ) {
       $job{W} = $job{'b'} - $st;
    } elsif ( $st > 0 ) {
       $job{W} = time() - $st;
    } else {
       $job{W} = "";
    }
    push @queue, \%job;
  }
  return \@queue;
}


sub queue_user {
# Get queue state using sinfo
# parameter user - filter by user
  my ($user) = @_; 
  my @res = ex("$config{PREFIX}squeue -h -o \"\%i;\%j;\%u\"");
  my @queue;
  foreach my $l (@res) {
    $l =~ s/\n//;
    my %job; my $j; my $u;
    ($job{'i'},$j, $u) = split(/;/,$l);
    $job{'j'} = "$j ($job{'i'})";
    push @queue, \%job if ($u eq $user);
  }
  return \@queue;
}


sub parse_log {
  my %j;

  foreach my $l (@_) {
    if ( $l =~ m/^$/) {
      next;
    }

    my ($id) = $l =~ m/JobId=(\d+)/;
    ($j{$id}{i}) = $l =~ m/JobId=(\d+)/;
    ($j{$id}{P}) = $l =~ m/Partition=([^ ]+)/;
    ($j{$id}{j}) = $l =~ m/Name=([^ ]+)/;
    ($j{$id}{u}) = $l =~ m/UserId=([^ \(\)]+)/;
    ($j{$id}{t}) = $l =~ m/JobState=([^ ]+)/;
    ($j{$id}{C}) = $l =~ m/ReqProcs=(\d+)/;
    ($j{$id}{C}) = $l =~ m/ReqNodes=(\d+)/;
    ($j{$id}{r}) = $l =~ m/Reason=([^ ]+)/;
    ($j{$id}{S}) = $l =~ m/SubmitTime=([^ ]+)/;
    ($j{$id}{b}) = $l =~ m/StartTime=([^ ]+)/;
    ($j{$id}{e}) = $l =~ m/EndTime=([^ ]+)/;
    ($j{$id}{E}) = $l =~ m/Dependency=([^ ]+)/;
    ($j{$id}{N}) = $l =~ m/NodeList=([^ ]+)/;

    # Postprocess
    $j{$id}{S} = 0 if not defined $j{$id}{S};
    $j{$id}{b} = 0 if not defined $j{$id}{b};
    $j{$id}{e} = 0 if not defined $j{$id}{e};

    $j{$id}{S} =~ s/-/ /;
    $j{$id}{b} =~ s/-/ /;
    $j{$id}{e} =~ s/-/ /;

    $j{$id}{S} = str2time($j{$id}{S});
    $j{$id}{b} = str2time($j{$id}{b});
    $j{$id}{e} = str2time($j{$id}{e});

    $j{$id}{S} = 0 if not defined $j{$id}{S};
    $j{$id}{b} = 0 if not defined $j{$id}{b};
    $j{$id}{e} = 0 if not defined $j{$id}{e};

    $j{$id}{M} = $j{$id}{e} - $j{$id}{b};
  
    $j{$id}{N} = slurm_expand_nodes($j{$id}{N});
  }
  return \%j;
}

sub parse_log_submittime {
  my %j;

  foreach my $l (@_) {
    if ( $l =~ m/^$/) {
      next;
    }
    my ($id) = $l =~ m/JobId=(\d+)/;
    ($j{$id}{S}) = $l =~ m/SubmitTime=([^ ]+)/;

    # Postprocess
    $j{$id}{S} = 0 if not defined $j{$id}{S};
    $j{$id}{S} =~ s/-/ /;
    $j{$id}{S} = str2time($j{$id}{S});
    $j{$id}{S} = 0 if not defined $j{$id}{S};
  }
  return \%j;
}

sub nodelist {
  my $hostlist_t = ex("$config{PREFIX}sinfo -h -t idle,alloc -o \"\%N\"");
  return split(',',slurm_expand_nodes($hostlist_t));
}

1;
