#!/usr/bin/perl -w
#
#   Copyright 2012 Andrii Golovynskyi
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

use Config::General;

use Cwd 'abs_path';
use File::Basename;

# Global variable 
our $prefix;
our $cmd_prefix;
our $cluster;
our $mpi_launcher;
our $queues;

my %config;

$ENV{LANG} = "C";

# Overriding config for remote cluster
if ((defined $cmd_prefix) && ($cmd_prefix ne "") &&
    (defined $cluster) && ($cluster ne "")) {
  $config{PREFIX} = "";
  $config{CLUSTER} = $cluster;
  $config{PBS_MPI_LAUNCHER} = $mpi_launcher if defined $mpi_launcher;
  $config{ALLOWED_QUEUES} = $queues if defined $queues;
} else {
  # Begin parsing config
  my $configfile = new Config::General("$prefix/etc/torque.conf");
  %config = $configfile->getall;
  $cmd_prefix = "";
  $config{PREFIX} .= '/';
}

use Date::Parse;

require "$prefix/libexec/modules/utils.pl";
require "$prefix/libexec/modules/makefile.pl";

# --------------------------- Functions ---------------------------------------------------

sub launch {
  my ($J,$d,$a,$p,$D,$Djid,$n,$t,$r,$g,$ppn) = @_;
  my $pbslauncher;
  if ( $r eq "mpi" ) {
    $pbslauncher= $config{PBS_MPI_LAUNCHER};
  } elsif ($r eq "none" ) {
    $pbslauncher= "";
  } else { 
   print ("Unknown Runtime: $r\n");
   exit;
  }

  # Running task

  system ("mkdir -p $d/$J.out");

  my $dependstring="";
  my $jobid="";

  my $prefix= $config{PREFIX};

#==============
  if (not defined $ppn) {
# TODO remote 
    $ppn = $config{PARTITION}->{"$p"}->{"PROCESSORS_PER_NODE"};
  }
  my $nodes;
  if( $n % $ppn == 0) {
    $nodes = ($n-$n % $ppn)/$ppn;
  } else {
    $nodes = ($n-$n % $ppn)/$ppn+1;
  }
  if ($D ne "") {
    $dependstring="-W depend=afterok:$Djid";
  }

  if ($D eq  "afterok") {
    $dependstring="-W depend=afterok:$Djid";
  } elsif ($D eq "afterfail") {
    $dependstring="-W depend=afterfail:$Djid";
  }

  my $cmd_str ="cd $d && echo \"$config{PBS_MPI_LAUNCHER} $a $g\" | $config{PREFIX}qsub -q $p $dependstring -d $d -l nodes=$nodes:ppn=$ppn,walltime=$t\:00 -N $J -o $d/$J.out/$J.out -e $d/$J.out/$J.err 2>&1";

  $res{cmd} = $cmd_str;
  $res{out} = ex($cmd_str);
  ($res{jobid}) = $res{out} =~ /([\d.\w]+)/;
  return \%res;
}

sub cancel {
# Cancel job "j", "a" -- administrator mode
  my ($j,$a) = @_;
  my $admprefix="";
  my $res;
 
  if ($a == 1 && ($cmd_prefix eq "")) {
    $admprefix="ssh root@".$config{ACCESS_SERVER};
  }
  $res = ex("$admprefix $config{PREFIX}qdel $j 2>&1");
  return $res;
}

sub resources {
# Get resources using pbsnodes
# parameter detail - detailed view
#
# TODO parse qstat -Q
#
  my ($detail) = @_;
  my @part_l;

  my @r = ex("$config{PREFIX}qstat -Q");
  @r = @r[2 .. $#r];

  foreach my $l (@r) {
    ($l) = $l =~ m/^([^ ]+)/;
    push @partitions, $l;
  }
  my @allowed_partitions = split(/,/,$config{TORQUE_ALLOWED_QUEUES});
  
#-------------------
  my $tr = ex("$config{PREFIX}pbsnodes");
  my @res = split(/\n\n/,$tr);

  my @idle;
  my @alloc;
  my @down;
  my @unknown;
  #cpu counts
  my $c_i=0; my $c_a=0; my $c_d=0; my $c_u=0;
  foreach my $i (@res) {
    my ($node) = $i =~ m/^(.+)/;
    my ($state) = $i =~ m/state = (.+)/;
    my ($ncpus) = $i =~ m/pcpus = (\d+)/;
    if ( $state =~ /free/ ) {
      push(@idle,$node);
      $c_i += $ncpus;
    } elsif ($state =~ /down/ ) {
       push(@down,$node);
      $c_d += $ncpus;
    } elsif ($state =~ /job-exclusive/ ) {
       push(@alloc,$node);
      $c_a += $ncpus;
    } else {
       push(@unknown,$node);
      $c_u += $ncpus;
    }
  }
  
  foreach my $p (@partitions) {
    $p =~ s/\n//;
    # idle
    if (@idle > 0) {
      my %p_i;
      $p_i{P} = $p;
      $p_i{t} = "idle";
      if (grep /$p/, @allowed_partitions) {
        $p_i{a} = "up";
      } else {
        $p_i{a} = "down";
      }
      $p_i{C} = $c_i;
      for (my $i=0; $i<=@idle-2; $i++) {
        $p_i{N} .= "$idle[$i],";
      }
      $p_i{N} .= "$idle[@idle-1]";
      $p_i{H} = $config{CLUSTER};
      $p_i{D} = $#idle+1;
      push @part_l, \%p_i;
    }
    # alloc
    if (@alloc > 0) {
      my %p_a;
      $p_a{P} = $p;
      $p_a{t} = "alloc";
      if (grep /$p/, @allowed_partitions) {
        $p_a{a} = "up";
      } else {
        $p_a{a} = "down";
      }
      $p_a{C} = $c_a;
      for (my $i=0; $i<=@alloc-2; $i++) {
        $p_a{N} .= "$alloc[$i],";
      }
      $p_a{N} .= "$alloc[@alloc-1]";
      $p_a{H} = $config{CLUSTER};
      $p_a{D} = $#alloc+1;
      push @part_l, \%p_a;
    }
    # down 
    if (@down > 0) {
      my %p_d;
      $p_d{P} = $p;
      $p_d{t} = "down";
      $p_d{a} = "down";
      $p_d{C} = $c_d;
      for (my $i=0; $i<=@down-2; $i++) {
        $p_d{N} .= "$down[$i],";
      }
      $p_d{N} .= "$down[@down-1]";
      $p_d{H} = $config{CLUSTER};
      $p_d{D} = $#down+1;
      push @part_l, \%p_d;
    }
    #unknown
    if (@unknown > 0) {
      my %p_u;
      $p_u{P} = $p;
      $p_u{t} = "unknown";
      $p_u{C} = $c_u;
      for (my $i=0; $i<=@unknown-2; $i++) {
        $p_u{N} .= "$unknown[$i],";
      }
      $p_u{N} .= "$unknown[@unknown-1]";
      $p_u{H} = $config{CLUSTER};
      $p_u{D} = $#unknown+1;
      push @part_l, \%p_u;
    }
 
  }
  return \@part_l;  
}

sub queue {
# Get queue state using qstat
# parameter detail - detailed view
  my ($detail) = @_; 
  my @res = ex("$config{PREFIX}qstat -f -1");
#---------------

# Torque variables
#ctime 	time job was created
#etime 	time job was queued
#qtime 	time job became eligible to run
#start 	time job started to run

  my @queue;
  my %job;

  foreach my $l (@res) {
    if (my ($id) = $l =~ m/Job Id: (.+)/){
      if ( %job) {
        my %j = %job;
        push @queue, \%j;
        %job = ();
      }
      $job{'i'} = $id;
      $job{'H'} = $config{CLUSTER};
      next;
    }
    if (my ($var,$val) = $l =~ m/ *(.+) = (.+)/){
      if ( $var eq "job_state") {
        $val = "PD" if $val eq "Q";
        $job{'t'} = $val;
      } elsif ($var eq "queue") {
        $job{'P'} = $val;
      } elsif ($var eq "ctime") {
        $job{'s'} = str2time($val);
      } elsif ($var eq "start_time") {
        $job{'b'} = str2time($val);
      } elsif ($var eq "etime") {
        $job{'e'} = str2time($val);
      } elsif ($var eq "Job_Owner") {
        ($job{'u'}) = $val =~ /(.+)\@/;
      } elsif ($var eq "Job_Name") {
        $job{'j'} = $val;
      } elsif ($var eq "exec_host") {
        my @c = split (/\+/,$val);
        $job{'C'} = $#c + 1;
        $val =~ s/\+/, /g;
        $job{'N'} = $val;
      } elsif ($var eq "resources_used.walltime") {
        $job{'M'} = str2sec($val);
      } elsif (($var eq "Resource_List.neednodes" || $var eq "Resource_List.nodes") && (not defined $job{'C'})) {
        $job{'C'} = 0;
        foreach my $l (split /\+/,$val) {
          ($n,$c) = $l =~ /(\d+)\:ppn=(\d+)/ if $l =~ /(\d+)\:ppn=(\d+)/;
          $job{'C'} += $n*$c if (defined $n && defined $c);
        }
      } else {
        $job{$var} = $val;
      }
      next;
    }
  }
  # push last job
  my %ttj = %job;
  push @queue, \%ttj;
  %job = ();
  return \@queue;
}


sub queue_user {
# Get only job IDs
# parameter user - filter by user
  my ($user) = @_;

  my @res = ex("$config{PREFIX}qstat -f -1");
  my @queue;
  my %job;

  foreach my $l (@res) {
    if (my ($id) = $l =~ m/Job Id: (.+)/){
      if ( %job) {
        my %j = %job;
        push @queue, \%j if ($job{'u'} == $user);
        %job = ();
      }
      $job{'i'} = $id;
      next;
    }
    if (my ($var,$val) = $l =~ m/ *(.+) = (.+)/){
      if ( $var eq "Job_Owner") {
        ($job{'u'}) = $val =~ /(.+)\@/;
      }
      next;
    }
  }

  # push last job
  my %ttj = %job;
  push @queue, \%ttj if ($job{'u'} == $user);
  %job = ();
  return \@queue;
}


# TODO
sub nodelist {
  my $hostlist_t = ex("$config{PREFIX}sinfo -h -t idle,alloc -o \"\%N\"");
  return split(',',slurm_expand_nodes($hostlist_t));
}

1;
