#!/usr/bin/perl -w

use strict;
use Getopt::Std;
use JSON;

my %options = ();
getopts("hLp:d:q:C:K:P:H:Q:M:U:W:D:",\%options);

sub help {
  print "Usage:
-p PATH     hydra installation path
-d PATH     testing directory path, test files will be created there
-q QUEUE    cluster partition or queue
-C CERT     grid certificate file
-P PASSW    grid certificate password
-K KEY      grid key
-H CLUSTER  grid cluster for job submission
-Q QUEUE    queue on remote grid cluster
-M HOST     mysql host (webserver)
-U USER     mysql user
-W PASSWD   mysql password
-D DATABASE mysql database name
-L          list test commands for individual testing of some function
-h          prints this help\n";
  exit;
}

help() if not $options{p};
help() if not $options{d};
help() if not $options{q};
help() if $options{h};


my $p = $options{p}."/libexec";

if ($> == 0) {
 print "Verification should be run from ordinary user account\n";
 exit 1;
}

sub tst {
  my ($n,$c) = @_;
  my $res = "";
  print "Checking $n\n";
  print "Command: $c\n";
  if (not defined $options{L}) {
    $res = `$c`;
    if ($res eq "") {
      print "Result: Error\n";
    } else {
      print "Result:\n";
      print "$res\n";
    }
  }
  return $res;
}

my @tests = (
# File manager
["user home", "echo '[{\"command\":\"home\"}]'| $p/hydra -p"],
["directory listing", "echo '[{\"command\":\"ls\", \"dir\":\"/home\"}]'| $p/hydra -p"],
["making directory", "echo '[{\"command\":\"mkdir\", \"dir\":\"$options{d}/created-dir\"}]'| $p/hydra -p"],
["Uploading file", "echo 'Hello Cluster\nThis is test file'| $p/hydra_io -c put -f \"$options{d}/testfile.txt\""],
["Downloading file", "$p/hydra_io -c get -f \"$options{d}/testfile.txt\""],
["Viewing file part of file", "echo '[{\"command\":\"less\", \"f\":\"$options{d}/testfile.txt\", \"o\":14, \"s\":20}]'| $p/hydra -p"],
# TODO Add files of all types, send them to testdir, unzip
["Archiving files", "echo '[{\"command\":\"zip\", \"dir\":\"$options{d}\", \"files\":[\"$options{d}/testfile.txt\", \"$options{d}/created-dir\"], \"a\":\"archive\"}]'| $p/hydra -p"],
["Creating file for removal", "echo 'Hello Cluster\nThis is removing test file'| $p/hydra_io -c put -f \"$options{d}/removefile.txt\""],
["file removing", "echo '[{\"command\":\"rm\", \"files\":[\"$options{d}/removefile.txt\"]}]'| $p/hydra -p"],
["directory copying", "echo '[{\"command\":\"cp\", \"source\":[\"$options{d}/created-dir\"],  \"target\":\"$options{d}/created-dir.copy\"}]'| $p/hydra -p"],
["making directory for movement", "echo '[{\"command\":\"mkdir\", \"dir\":\"$options{d}/dir-to-move\"}]'| $p/hydra -p"],
["directory moving", "echo '[{\"command\":\"mv\", \"source\":[\"$options{d}/dir-to-move\"],  \"target\":\"$options{d}/dir-moved\"}]'| $p/hydra -p"],
["changing file permissions", "echo '[{\"command\":\"chmod\", \"files\":[\"$options{d}/created-dir\"],  \"mode\":\"777\", \"recursive\":0}]'| $p/hydra -p"],
["directory size", "echo '[{\"command\":\"du\", \"files\":[\"$options{d}\"]}]'| $p/hydra -p"],

# Resource manager
["cluster resources list", "echo '[{\"command\":\"resources\"}]'| $p/hydra -p"],
["cluster task queue list", "echo '[{\"command\":\"queues\", \"d\":1}]'| $p/hydra -p"],
["cluster task subsission", "echo '[{\"command\":\"launch\", \"J\":\"hostname\", \"d\":\"$options{d}\", \"r\":\"mpi\", \"a\":\"/bin/hostname\", \"p\":\"$options{q}\", \"n\":\"1\", \"t\":\"5\", \"m\":0, \"g\":\"\", \"D\":\"\"}]'| $p/hydra -p"],
["staring long cluster task", "echo '[{\"command\":\"launch\", \"J\":\"sleep10min\", \"d\":\"$options{d}\", \"r\":\"mpi\", \"a\":\"/bin/sleep\", \"p\":\"$options{q}\", \"n\":\"1\", \"t\":\"20\", \"m\":0, \"g\":\"600\", \"D\":\"\"}]'| $p/hydra -p"]
# TODO check canceling task
);

# TODO check file permissions

my $res;

foreach my $t (@tests) {
  tst(@$t);
}

# Grid middleware
if (defined $options{C}) {
  if ((not defined $options{K}) || (not defined $options{P}) || (not defined $options{H}) || (not defined $options{Q})
    || (not defined $options{M}) || (not defined $options{U}) || (not defined $options{W}) || (not defined $options{D}) ) {
    print "Warning: For grid tests you should define key file and password, grid cluster, queue, mysql parameters\n";
    exit 1;
  }
  my $c = `cat $options{C}`;
  $c =~ s/\n/\\n/gi;
  my $k = `cat $options{K}`;
  $k =~ s/\n/\\n/gi;


  tst(("installation of grid certificates", "echo '[{\"command\":\"grid_cert_install\", \"cert\":\"$c\", \"key\":\"$k\"}]'| $p/hydra -p"));
  tst(("grid certificate info", "echo '[{\"command\":\"grid_cert_diag\"}]'| $p/hydra -p"));
  tst(("initializing grid proxy", "echo '[{\"command\":\"grid_proxy_init\", \"p\":\"$options{P}\"}]'| $p/hydra -p"));
  tst(("grid proxy info", "echo '[{\"command\":\"grid_proxy_info\"}]'| $p/hydra -p"));
  tst(("grid resources list", "echo '[{\"command\":\"grid_resources\"}]'| $p/hydra -p"));
  $res = tst(("grid task submission", "echo '[{\"command\":\"grid_launch\", \"H\":\"$options{H}\", \"J\":\"hostname\", \"d\":\"$options{d}\", \"iF\":\"\", \"oF\":\"\", \"a\":\"/bin/hostname\", \"e\":\"\", \"p\":\"$options{Q}\", \"n\": 1, \"t\":10, \"r\":\"custom\", \"mysql_host\":\"$options{M}\", \"mysql_login\":\"$options{U}\", \"mysql_pass\":\"$options{W}\", \"mysql_db\":\"$options{D}\" }]'| $p/hydra -p"));
  my $json = JSON->new->allow_nonref;
  my $r = $json->decode($res);
  my $jobid = $r->{arc_launch}{id};
  print "Wait a little grid task to start\n";
  sleep 20; 

  tst(("grid task queue listing", "echo '[{\"command\":\"grid_queue\"}]'| $p/hydra -p"));
  tst(("directory listing with remote dirs of grid tasks", "echo '[{\"command\":\"ls\", \"dir\":\"$options{d}\",  \"mysql_host\":\"$options{M}\", \"mysql_login\":\"$options{U}\", \"mysql_pass\":\"$options{W}\", \"mysql_db\":\"$options{D}\"}]'| $p/hydra -p"));
  tst(("grid task directory listing", "echo '[{\"command\":\"ls\", \"m\":\"grid\", \"dir\":\"$jobid\",  \"mysql_host\":\"$options{M}\", \"mysql_login\":\"$options{U}\", \"mysql_pass\":\"$options{W}\", \"mysql_db\":\"$options{D}\"}]'| $p/hydra -p"));
  tst(("Viewing file part of grid task file hostname.out", "echo '[{\"command\":\"less\", \"m\":\"grid\", \"f\":\"$jobid/hostname.out\", \"o\":14, \"s\":20}]'| $p/hydra -p"));
  tst(("Viewing file part of grid task file hostname.err", "echo '[{\"command\":\"less\", \"m\":\"grid\", \"f\":\"$jobid/hostname.err\", \"o\":14, \"s\":20}]'| $p/hydra -p"));
  tst(("grid task getting files", "echo '[{\"command\":\"grid_cp\", \"sf\":[\"$jobid/hostname.err\", \"$jobid/hostname.out\"], \"ddir\":\"$options{d}\"}]'| $p/hydra -p"));
  tst(("grid task putting files", "echo '[{\"command\":\"grid_cp\", \"sf\":[\"$options{d}/testfile.txt\"], \"ddir\":\"$jobid\"}]'| $p/hydra -p"));
  tst(("cancelling grid task", "echo '[{\"command\":\"grid_cancel\", \"jid\":\"$jobid\"}]'| $p/hydra -p"));
  tst(("making directory for movement", "echo '[{\"command\":\"mkdir\", \"dir\":\"$options{d}/grid-task\"}]'| $p/hydra -p"));
  tst(("downloading completed grid task", "echo '[{\"command\":\"grid_get\", \"jid\":\"$jobid\", \"dir\":\"$options{d}/grid-task\"}]'| $p/hydra -p"));
  tst(("directory listing", "echo '[{\"command\":\"ls\", \"dir\":\"$options{d}/grid-task\"}]'| $p/hydra -p"));

  print "Run as root /opt/hydra/libexec/monitoring/grid_jobget-cron to check downloading of completed tasks\n";
  tst(("grid proxy destroy", "echo '[{\"command\":\"grid_proxy_destroy\"}]'| $p/hydra -p"));
}
